"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Differ = void 0;
const assert_1 = __importDefault(require("assert"));
const console_1 = __importDefault(require("console"));
const chalk_1 = __importDefault(require("chalk"));
const DiffItem_1 = require("./DiffItem");
const diffStringSet = (set1, set2) => {
    const added = new Set();
    const removed = new Set();
    const stable = new Set();
    for (const s of set1) {
        if (set2.has(s)) {
            stable.add(s);
        }
        else {
            removed.add(s);
        }
    }
    for (const s of set2) {
        if (!set1.has(s)) {
            added.add(s);
        }
    }
    return [added, removed, stable];
};
const diffMapKeys = (map1, map2) => {
    const keys1 = new Set(map1.keys());
    const keys2 = new Set(map2.keys());
    return diffStringSet(keys1, keys2);
};
class Differ {
    #model1;
    #model2;
    #items;
    #outdatedVersion;
    #newVersion;
    constructor(model1, model2) {
        this.#model1 = model1;
        this.#model2 = model2;
        this.#items = [];
        this.#outdatedVersion = false;
        this.#newVersion = this.#model2.info.version;
    }
    calculate() {
        this.diffInfo();
        this.diffServers();
        this.diffPaths();
        this.diffComponents();
        this.diffSecurityRequirements();
        this.diffTags();
        this.diffExternalDocs(this.#model1.externalDocs, this.#model2.externalDocs, '#');
        this.calculateNewVersion();
    }
    diffInfo() {
        const info1 = this.#model1.info;
        const info2 = this.#model2.info;
        // x-id is a special extension, check it first
        const id1 = info1.getExtension('id');
        const id2 = info2.getExtension('id');
        (0, assert_1.default)(id1 === id2, `Expected #/info/x-id to be identical, but they were not`);
        if (info1.title !== info2.title) {
            this.#items.push(new DiffItem_1.DiffItem('#/info/title', 'major', 'changed'));
        }
        if (info1.description !== info2.description) {
            this.#items.push(new DiffItem_1.DiffItem('#/info/description', 'patch', 'changed'));
        }
        if (info1.termsOfService !== info2.termsOfService) {
            this.#items.push(new DiffItem_1.DiffItem('#/info/termsOfService', 'minor', 'changed'));
        }
        // skip info.version, because it is our output, not input
        // rest of extension fields
        const audience1 = info1.getExtension('audience');
        const audience2 = info2.getExtension('audience');
        if (audience1 !== audience2) {
            this.#items.push(new DiffItem_1.DiffItem('#/info/x-audience', 'major', 'changed'));
        }
        const rootUrl1 = info1.getExtension('root-url');
        const rootUrl2 = info2.getExtension('root-url');
        if (rootUrl1 !== rootUrl2) {
            this.#items.push(new DiffItem_1.DiffItem('#/info/x-root-url', 'minor', 'changed'));
        }
        this.diffContact();
        this.diffLicense();
    }
    diffContact() {
        const contact1 = this.#model1.info.contact;
        const contact2 = this.#model2.info.contact;
        if (contact1.name !== contact2.name) {
            this.#items.push(new DiffItem_1.DiffItem('#/info/contact/name', 'patch', `changed`));
        }
        if (contact1.url !== contact2.url) {
            this.#items.push(new DiffItem_1.DiffItem('#/info/contact/url', 'patch', 'changed'));
        }
        if (contact1.email !== contact2.email) {
            this.#items.push(new DiffItem_1.DiffItem('#/info/contact/email', 'patch', 'changed'));
        }
    }
    diffLicense() {
        const license1 = this.#model1.info.license;
        const license2 = this.#model2.info.license;
        if (license1.name !== license2.name) {
            this.#items.push(new DiffItem_1.DiffItem('#/info/license/name', 'minor', 'changed'));
        }
        if (license1.url !== license2.url) {
            this.#items.push(new DiffItem_1.DiffItem('#/info/license/url', 'minor', 'changed'));
        }
    }
    diffServers() {
        const servers1 = this.#model1.servers;
        const servers2 = this.#model2.servers;
        const len = Math.max(servers1.length, servers2.length);
        for (let i = 0; i < len; i += 1) {
            const server1 = servers1[i];
            const server2 = servers2[i];
            if (server1 == null && server2 != null) {
                this.#items.push(new DiffItem_1.DiffItem(`#/servers/${i}`, 'minor', `added`));
            }
            else if (server1 != null && server2 == null) {
                this.#items.push(new DiffItem_1.DiffItem(`#/servers/${i}`, 'minor', `removed`));
            }
            else {
                if (server1.url !== server2.url) {
                    this.#items.push(new DiffItem_1.DiffItem(`#/servers/${i}/url`, 'minor', `changed`));
                }
                if (server1.description !== server2.description) {
                    this.#items.push(new DiffItem_1.DiffItem(`#/servers/${i}/description`, 'patch', 'changed'));
                }
            }
        }
    }
    diffPaths() {
        const paths1 = this.#model1.paths;
        const paths2 = this.#model2.paths;
        const urls1 = new Set(paths1.keys());
        const urls2 = new Set(paths2.keys());
        const allUrls = new Set([...urls1, ...urls2]);
        for (const url of allUrls) {
            if (urls1.has(url) && !urls2.has(url)) {
                this.#items.push(new DiffItem_1.DiffItem(`#/paths${url}`, 'major', `removed`));
            }
            else if (!urls1.has(url) && urls2.has(url)) {
                this.#items.push(new DiffItem_1.DiffItem(`#/paths${url}`, 'minor', `added`));
            }
            else {
                this.diffPathItem(url);
            }
        }
    }
    diffPathItem(pathUrl) {
        const item1 = this.#model1.paths.getItemOrThrow(pathUrl);
        const item2 = this.#model2.paths.getItemOrThrow(pathUrl);
        if (item1.summary !== item2.summary) {
            this.#items.push(new DiffItem_1.DiffItem(`#/paths${pathUrl}/summary`, 'minor', `changed`));
        }
        if (item1.description !== item2.description) {
            this.#items.push(new DiffItem_1.DiffItem(`#/paths${pathUrl}/description`, 'patch', `changed`));
        }
        this.diffParameters(item1.parameters, item2.parameters, `#/paths${pathUrl}/parameters`);
        const httpMethods = [
            'get',
            'put',
            'post',
            'delete',
            'options',
            'head',
            'patch',
            'trace',
        ];
        for (const httpMethod of httpMethods) {
            const operation1 = item1.getOperation(httpMethod);
            const operation2 = item2.getOperation(httpMethod);
            if (operation2 != null) {
                if (operation1 != null) {
                    this.diffOperation(pathUrl, httpMethod);
                }
                else {
                    this.#items.push(new DiffItem_1.DiffItem(`#/paths${pathUrl}/${httpMethod}`, 'minor', `operation added`));
                }
            }
            else if (operation1 != null) {
                this.#items.push(new DiffItem_1.DiffItem(`#/paths${pathUrl}/${httpMethod}`, 'major', `operation removed`));
            }
        }
    }
    diffParameters(parameters1, parameters2, basePointer) {
        const params1 = new Map(parameters1.map((p, i) => [`${p.in}:${p.name}`, [p, i]]));
        const params2 = new Map(parameters2.map((p, i) => [`${p.in}:${p.name}`, [p, i]]));
        const [added, removed, stable] = diffStringSet(new Set(params1.keys()), new Set(params2.keys()));
        for (const key of added) {
            const p = params2.get(key);
            (0, assert_1.default)(p);
            this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/${p[1]}`, p[0].required ? 'major' : 'minor', `added ${key}`));
        }
        for (const key of removed) {
            const p = params1.get(key);
            (0, assert_1.default)(p);
            this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/${p[1]}`, 'major', `removed ${key}`));
        }
        for (const key of stable) {
            const param1 = params1.get(key);
            (0, assert_1.default)(param1);
            const param2 = params2.get(key);
            (0, assert_1.default)(param2);
            this.diffParameter(param1, param2, basePointer);
        }
    }
    diffParameter([param1], [param2, index2], basePointer) {
        if (param1.in !== param2.in) {
            this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/${index2}/in`, 'major', 'changed'));
        }
        if (param1.name !== param2.name) {
            this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/${index2}/name`, 'major', 'changed'));
        }
        if (param1.description !== param2.description) {
            this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/${index2}/description`, 'patch', 'changed'));
        }
        if (param1.deprecated !== param2.deprecated) {
            this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/${index2}/deprecated`, 'major', 'changed'));
        }
        if (param1.explode !== param2.explode) {
            this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/${index2}/explode`, 'major', 'changed'));
        }
        // schema
        this.diffSchema(param1.schema, param2.schema, `${basePointer}/${index2}/schema`);
        // example, examples - skip, because it does not affect version
        // content
    }
    diffSchema(schema1, schema2, basePointer) {
        if (schema2 != null) {
            if (schema1 != null) {
                if (schema1.title !== schema2.title) {
                    this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/title`, 'minor', 'changed'));
                }
                if (schema1.multipleOf !== schema2.multipleOf) {
                    this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/multipleOf`, 'major', 'changed'));
                }
                if (schema1.maximum !== schema2.maximum) {
                    this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/maximum`, 'major', 'changed'));
                }
                if (schema1.exclusiveMaximum !== schema2.exclusiveMaximum) {
                    this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/exclusiveMaximum`, 'major', 'changed'));
                }
                if (schema1.minimum !== schema2.minimum) {
                    this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/minimum`, 'major', 'changed'));
                }
                if (schema1.exclusiveMinimum !== schema2.exclusiveMinimum) {
                    this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/exclusiveMinimum`, 'major', 'changed'));
                }
                if (schema1.maxLength !== schema2.maxLength) {
                    this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/maxLength`, 'major', 'changed'));
                }
                if (schema1.minLength !== schema2.minLength) {
                    this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/minLength`, 'major', 'changed'));
                }
                if (schema1.pattern !== schema2.pattern) {
                    this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/pattern`, 'major', 'changed'));
                }
                if (schema1.minItems !== schema2.minItems) {
                    this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/minItems`, 'major', 'changed'));
                }
                if (schema1.maxItems !== schema2.maxItems) {
                    this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/maxItems`, 'major', 'changed'));
                }
                if (schema1.uniqueItems !== schema2.uniqueItems) {
                    this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/uniqueItems`, 'major', 'changed'));
                }
                if (schema1.maxProperties !== schema2.maxProperties) {
                    this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/maxProperties`, 'major', 'changed'));
                }
                if (schema1.minProperties !== schema2.minProperties) {
                    this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/minProperties`, 'major', 'changed'));
                }
                const [addedRequired, removedRequired] = diffStringSet(schema1.required, schema2.required);
                for (const p of addedRequired) {
                    this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/required`, 'major', `added ${p}`));
                }
                for (const p of removedRequired) {
                    this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/required`, 'major', `removed ${p}`));
                }
                if (schema1.type !== schema2.type) {
                    this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/type`, 'major', `changed`));
                }
                this.diffSchema(schema1.items, schema2.items, `${basePointer}/items`);
                const propNames1 = new Set(schema1.properties.keys());
                const propNames2 = new Set(schema2.properties.keys());
                const [addedProps, removedProps, stableProps] = diffStringSet(propNames1, propNames2);
                for (const p of addedProps) {
                    this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/properties/${p}`, 'minor', 'added'));
                }
                for (const p of removedProps) {
                    this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/properties/${p}`, 'major', 'removed'));
                }
                for (const p of stableProps) {
                    const subschema1 = schema1.getPropertyOrThrow(p);
                    const subschema2 = schema2.getPropertyOrThrow(p);
                    this.diffSchema(subschema1, subschema2, `${basePointer}/properties/${p}`);
                }
                if (schema1.description !== schema2.description) {
                    this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/description`, 'patch', 'changed'));
                }
                if (schema1.format !== schema2.format) {
                    this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/format`, 'major', 'changed'));
                }
                if (schema1.default !== schema2.default) {
                    this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/default`, 'major', 'changed'));
                }
                if (schema1.nullable !== schema2.nullable) {
                    this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/nullable`, 'major', 'changed'));
                }
                if (schema1.readOnly !== schema2.readOnly) {
                    this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/readOnly`, 'patch', 'changed'));
                }
                if (schema1.writeOnly !== schema2.writeOnly) {
                    this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/writeOnly`, 'patch', 'changed'));
                }
                if (schema1.deprecated !== schema2.deprecated) {
                    this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/deprecated`, 'major', 'deprecated'));
                }
            }
            else {
                this.#items.push(new DiffItem_1.DiffItem(basePointer, 'major', 'added'));
            }
        }
        else if (schema1 != null) {
            this.#items.push(new DiffItem_1.DiffItem(basePointer, 'major', 'removed'));
        }
    }
    diffOperation(pathUrl, httpMethod) {
        const operation1 = this.#model1.paths.getItemOrThrow(pathUrl).getOperationOrThrow(httpMethod);
        const operation2 = this.#model2.paths.getItemOrThrow(pathUrl).getOperationOrThrow(httpMethod);
        // tags
        const tags1 = new Set(operation1.tags);
        const tags2 = new Set(operation2.tags);
        const [added, removed] = diffStringSet(tags1, tags2);
        for (const t of added) {
            this.#items.push(new DiffItem_1.DiffItem(`#/paths${pathUrl}/${httpMethod}/tags`, 'patch', `added ${t}`));
        }
        for (const t of removed) {
            this.#items.push(new DiffItem_1.DiffItem(`#/paths${pathUrl}/${httpMethod}/tags`, 'patch', `removed ${t}`));
        }
        if (operation1.summary !== operation2.summary) {
            this.#items.push(new DiffItem_1.DiffItem(`#/paths${pathUrl}/${httpMethod}/summary`, 'minor', `changed`));
        }
        if (operation1.description !== operation2.description) {
            this.#items.push(new DiffItem_1.DiffItem(`#/paths${pathUrl}/${httpMethod}/description`, 'patch', `changed`));
        }
        // this.diffExternalDocs();
        if (operation1.operationId !== operation2.operationId) {
            this.#items.push(new DiffItem_1.DiffItem(`#/paths${pathUrl}/${httpMethod}/operationId`, 'major', `changed`));
        }
        // parameters
        this.diffParameters(operation1.parameters, operation2.parameters, `#/paths${pathUrl}/${httpMethod}/`);
        // this.diffRequestBody();
        // this.diffResponses();
        // this.diffCallbacks();
        if (operation1.deprecated !== operation2.deprecated) {
            this.#items.push(new DiffItem_1.DiffItem(`#/paths${pathUrl}/${httpMethod}/deprecated`, 'major', `changed`));
        }
    }
    diffComponents() {
        const components1 = this.#model1.components;
        const components2 = this.#model2.components;
        // stable schemas (and other shared objects) should be compared inside other methods, like diffOperation
        this.diffMapKeys(components1.schemas, components2.schemas, '#/components/schemas');
        this.diffMapKeys(components1.responses, components2.responses, '#/components/responses');
        this.diffMapKeys(components1.parameters, components2.parameters, '#/components/parameters');
        this.diffMapKeys(components1.examples, components2.examples, '#/components/examples');
        this.diffMapKeys(components1.requestBodies, components2.requestBodies, '#/components/requestBodies');
        this.diffMapKeys(components1.headers, components2.headers, '#/components/headers');
        this.diffMapKeys(components1.securitySchemes, components2.securitySchemes, '#/components/securitySchemes');
        this.diffMapKeys(components1.links, components2.links, '#/components/links');
        this.diffMapKeys(components1.callbacks, components2.callbacks, '#/components/callbacks');
    }
    diffMapKeys(map1, map2, basePointer) {
        const keys1 = new Set(map1.keys());
        const keys2 = new Set(map2.keys());
        const [added, removed, stable] = diffStringSet(keys1, keys2);
        for (const k of added) {
            this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/${k}`, 'minor', 'added'));
        }
        for (const k of removed) {
            this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/${k}`, 'minor', 'removed'));
        }
        return [added, removed, stable];
    }
    diffSecurityRequirements() {
        const aggregatedReqs1 = new Map();
        const aggregatedReqs2 = new Map();
        for (const req of this.#model1.security) {
            for (const [schemeName, scopes] of req.scopes) {
                const aggregatedScopes = aggregatedReqs1.get(schemeName);
                if (aggregatedScopes) {
                    for (const s of scopes) {
                        aggregatedScopes.add(s);
                    }
                }
                else {
                    aggregatedReqs1.set(schemeName, new Set(scopes));
                }
            }
        }
        for (const req of this.#model1.security) {
            for (const [schemeName, scopes] of req.scopes) {
                const aggregatedScopes = aggregatedReqs2.get(schemeName);
                if (aggregatedScopes) {
                    for (const s of scopes) {
                        aggregatedScopes.add(s);
                    }
                }
                else {
                    aggregatedReqs2.set(schemeName, new Set(scopes));
                }
            }
        }
        const [keysAdded, keysRemoved, keysStable] = diffMapKeys(aggregatedReqs1, aggregatedReqs2);
        for (const k of keysAdded) {
            this.#items.push(new DiffItem_1.DiffItem('#/security', 'minor', `added scheme ${k}`));
        }
        for (const k of keysRemoved) {
            this.#items.push(new DiffItem_1.DiffItem('#/security', 'minor', `removed scheme ${k}`));
        }
        for (const k of keysStable) {
            const [scopesAdded, scopesRemoved] = diffStringSet(aggregatedReqs1.get(k), aggregatedReqs2.get(k));
            for (const s of scopesAdded) {
                this.#items.push(new DiffItem_1.DiffItem('#/security', 'minor', `added scope ${s} to scheme ${k}`));
            }
            for (const s of scopesRemoved) {
                this.#items.push(new DiffItem_1.DiffItem('#/security', 'minor', `removed scope ${s} from scheme ${k}`));
            }
        }
    }
    diffTags() {
        const tags1 = this.#model1.tags;
        const tags2 = this.#model2.tags;
        const tags1ByName = new Map(tags1.map((t, i) => [t.name, [t, i]]));
        const tags2ByName = new Map(tags2.map((t, i) => [t.name, [t, i]]));
        const [addedTags, removedTags, stableTags] = diffMapKeys(tags1ByName, tags2ByName);
        for (const name of addedTags) {
            const entry2 = tags2ByName.get(name);
            (0, assert_1.default)(entry2);
            this.#items.push(new DiffItem_1.DiffItem(`#/tags/${entry2[1]}`, 'minor', 'added'));
        }
        for (const name of removedTags) {
            const entry2 = tags2ByName.get(name);
            (0, assert_1.default)(entry2);
            this.#items.push(new DiffItem_1.DiffItem(`#/tags/${entry2[1]}`, 'minor', 'removed'));
        }
        for (const name of stableTags) {
            const entry1 = tags1ByName.get(name);
            (0, assert_1.default)(entry1);
            const entry2 = tags2ByName.get(name);
            (0, assert_1.default)(entry2);
            this.diffTag(entry1[0], entry2[0], `#/tags/${entry2[1]}`);
        }
    }
    diffTag(tag1, tag2, basePointer) {
        if (tag1.name !== tag2.name) {
            this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/name`, 'minor', 'changed'));
        }
        if (tag1.description !== tag2.description) {
            this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/description`, 'patch', 'changed'));
        }
        this.diffExternalDocs(tag1.externalDocs, tag2.externalDocs, '#/');
    }
    diffExternalDocs(docs1, docs2, basePointer) {
        if (docs2 != null) {
            if (docs1 != null) {
                if (docs1.url !== docs2.url) {
                    this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/url`, 'patch', 'changed'));
                }
                if (docs1.description !== docs2.description) {
                    this.#items.push(new DiffItem_1.DiffItem(`${basePointer}/description`, 'patch', 'changed'));
                }
            }
            else {
                this.#items.push(new DiffItem_1.DiffItem(basePointer, 'patch', 'added'));
            }
        }
        else if (docs1 != null) {
            this.#items.push(new DiffItem_1.DiffItem(basePointer, 'patch', 'removed'));
        }
    }
    calculateNewVersion() {
        const severities = new Set();
        for (const item of this.#items) {
            severities.add(item.severity);
        }
        let [major, minor, patch] = this.#model2.info.version.split('.');
        if (severities.has('major')) {
            major = `${Number(major) + 1}`;
        }
        else if (severities.has('minor')) {
            minor = `${Number(minor) + 1}`;
        }
        else if (severities.has('patch')) {
            patch = `${Number(patch) + 1}`;
        }
        this.#outdatedVersion = severities.size > 0;
        this.#newVersion = `${major}.${minor}.${patch}`;
    }
    print() {
        for (const item of this.#items) {
            let prefix;
            switch (item.severity) {
                case 'major':
                    prefix = chalk_1.default.red(item.severity);
                    break;
                case 'minor':
                    prefix = chalk_1.default.magenta(item.severity);
                    break;
                case 'patch':
                    prefix = chalk_1.default.yellow(item.severity);
                    break;
                default:
                    // eslint-disable-next-line @typescript-eslint/restrict-template-expressions
                    assert_1.default.fail(`Unsupported severity ${item.severity}`);
            }
            console_1.default.log(`[${prefix}] ${item.pointer} ${item.message}`);
        }
    }
    get outdatedVersion() {
        return this.#outdatedVersion;
    }
    get newVersion() {
        return this.#newVersion;
    }
}
exports.Differ = Differ;
//# sourceMappingURL=Differ.js.map